def import_mitsuba_backend():
    from tal.config import ask_for_config, Config
    mitsuba_version = ask_for_config(Config.MITSUBA_VERSION, force_ask=False)
    if mitsuba_version == '2':
        import tal.render.mitsuba2_transient_nlos as mitsuba_backend
    elif mitsuba_version == '3':
        import tal.render.mitsuba3_transient_nlos as mitsuba_backend
    else:
        raise AssertionError(
            f'Invalid MITSUBA_VERSION={mitsuba_version}, must be one of (2, 3)')
    return mitsuba_backend


def get_grid_xyz(nx, ny, rw_scale_x, rw_scale_y, ax0=0, ax1=1, ay0=0, ay1=1):
    import numpy as np
    px0 = -rw_scale_x + 2 * rw_scale_x * ax0
    px1 = rw_scale_x - 2 * rw_scale_x * (1 - ax1)
    py0 = -rw_scale_y + 2 * rw_scale_y * ay0
    py1 = rw_scale_y - 2 * rw_scale_y * (1 - ay1)
    xg = np.stack(
        (np.linspace(px0, px1, num=2*nx + 1)[1::2],)*ny, axis=1)
    yg = np.stack(
        (np.linspace(py0, py1, num=2*ny + 1)[1::2],)*nx, axis=0)
    assert xg.shape[0] == yg.shape[0] == nx and xg.shape[1] == yg.shape[1] == ny, \
        'Incorrect shapes'
    return np.stack([xg, yg, np.zeros((nx, ny))], axis=-1).astype(np.float32)


def expand_xy_dims(vec, x, y):
    assert len(vec) == 3
    return vec.reshape(1, 1, 3).repeat(x, axis=0).repeat(y, axis=1)



import os
import shutil
import yaml
import tal
import platform
from tal.io.capture_data import NLOSCaptureData
from tal.enums import FileFormat, GridFormat, HFormat, GroundTruthFormat
from tal.config import local_file_path
from tal.log import log, LogLevel, TQDMLogRedirect
from tal.render.util import get_grid_xyz, expand_xy_dims
import datetime
import numpy as np
from tqdm import tqdm
import multiprocessing
from functools import partial

from tal.render.util import import_mitsuba_backend


def _read_config_and_init_mitsuba_variant(config_path, args):
    # FIXME: on macOS "spawn" method, which is the default since 3.8,
    # is considered more safe than "fork", but requires serialization methods available
    # to send the objects to the spawned process. So a proper fix would be to add them
    # (see e.g. https://stackoverflow.com/a/65513291 and
    # https://docs.python.org/3/library/multiprocessing.html#the-spawn-and-forkserver-start-methods
    # for more details)
    if platform.system() == 'Darwin':
        multiprocessing.set_start_method('fork')

    config_path = os.path.abspath(config_path)

    assert os.path.exists(config_path), \
        f'{config_path} does not exist'

    path, name = os.path.split(config_path)
    candidate_filename = os.path.join(path, name, f'{name}.yaml')
    if os.path.isdir(config_path) and os.path.isfile(candidate_filename):
        config_path = candidate_filename

    assert os.path.isfile(config_path), \
        f'{config_path} is not a TAL config file'

    try:
        scene_config = yaml.safe_load(
            open(config_path, 'r')) or dict()

        if scene_config.get('mitsuba_variant', '').startswith('cuda'):
            assert len(args.gpus) > 0, \
                'You must specify at least one GPU to use CUDA. Use tal --gpu <id1> <id2> ...'
            gpu_ids = ','.join(map(str, args.gpus))
            os.environ['CUDA_VISIBLE_DEVICES'] = gpu_ids
        else:
            os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

        mitsuba_backend = import_mitsuba_backend()

        scene_defaults = yaml.safe_load(
            open(local_file_path('render/scene_defaults.yaml'), 'r'))
        scene_defaults['mitsuba_variant'] = mitsuba_backend.get_default_variant()

    except yaml.YAMLError as exc:
        raise AssertionError(
            f'Invalid YAML format in TAL config file: {exc}') from exc
    scene_config = {**scene_defaults, **scene_config}
    if not args.dry_run:
        mitsuba_backend.set_variant(scene_config['mitsuba_variant'])

    return mitsuba_backend, scene_config, config_path


def _check_progress_and_create_folders(config_path, args):
    config_dir, config_filename = os.path.split(config_path)
    try:
        in_progress = False
        progress_file = os.path.join(config_dir, 'IN_PROGRESS')
        if os.path.exists(progress_file):
            with open(progress_file, 'r') as f:
                progress_folder = f.read()
            if os.path.exists(os.path.join(config_dir, progress_folder)):
                in_progress = True
            else:
                log(LogLevel.INFO, 'The IN_PROGRESS file is stale, removing it...')
                os.remove(progress_file)
        if in_progress:
            log(LogLevel.INFO,
                f'Found a render in progress ({progress_folder}), continuing...')
        if not in_progress:
            progress_folder = datetime.datetime.now().strftime(r'%Y%m%d-%H%M%S')

        root_dir = os.path.join(config_dir, progress_folder)
        partial_results_dir = os.path.join(root_dir, 'partial')
        steady_dir = os.path.join(root_dir, 'steady')
        log_dir = os.path.join(root_dir, 'logs')

        if not in_progress:
            os.mkdir(root_dir)
            os.mkdir(partial_results_dir)
            os.mkdir(steady_dir)
            os.mkdir(log_dir)
            shutil.copy(
                config_path,
                os.path.join(root_dir, f'{config_filename}.old'))
        with open(progress_file, 'w') as f:
            f.write(progress_folder)
    except OSError as exc:
        raise AssertionError(f'Invalid permissions: {exc}') from exc

    return root_dir, partial_results_dir, steady_dir, log_dir, progress_file


def __write_scene_xmls(args, mitsuba_backend, scene_config, root_dir):
    steady_xml, ground_truth_xml, nlos_xml = mitsuba_backend.get_scene_xml(
        scene_config, random_seed=args.seed)

    steady_scene_xml = os.path.join(root_dir, 'steady_scene.xml')
    with open(steady_scene_xml, 'w') as f:
        f.write(steady_xml)

    ground_truth_scene_xml = os.path.join(
        root_dir, 'ground_truth_scene.xml')
    with open(ground_truth_scene_xml, 'w') as f:
        f.write(ground_truth_xml)

    nlos_scene_xml = os.path.join(root_dir, 'nlos_scene.xml')
    with open(nlos_scene_xml, 'w') as f:
        f.write(nlos_xml)

    return steady_scene_xml, ground_truth_scene_xml, nlos_scene_xml


def __write_metadata_and_get_laser_lookats(args, scene_config):
    """ Compute laser_lookats """

    laser_lookats = []
    scan_type = scene_config['scan_type']
    num_bins = scene_config['num_bins']
    sensor_width = scene_config['sensor_width']
    sensor_height = scene_config['sensor_height']
    laser_width = scene_config['laser_width']
    laser_height = scene_config['laser_height']
    if scan_type == 'single':
        laser_width = 1
        laser_height = 1

    relay_wall = next(filter(
        lambda g: g['name'] == scene_config['relay_wall'],
        scene_config['geometry']))
    assert 'rot_degrees_x' not in relay_wall and \
        'rot_degrees_y' not in relay_wall and \
        'rot_degrees_z' not in relay_wall, \
        'Relay wall displacement/rotation is NYI'

    laser_aperture_start_x = scene_config['laser_aperture_start_x'] or 0
    laser_aperture_start_y = scene_config['laser_aperture_start_y'] or 0
    laser_aperture_end_x = scene_config['laser_aperture_end_x'] or 1
    laser_aperture_end_y = scene_config['laser_aperture_end_y'] or 1

    if scan_type == 'single':
        laser_lookat_x = \
            scene_config['laser_lookat_x'] or sensor_width / 2
        laser_lookat_y = \
            scene_config['laser_lookat_y'] or sensor_height / 2
        laser_lookats.append((laser_lookat_x, laser_lookat_y))
    elif scan_type == 'exhaustive' or scan_type == 'confocal':
        assert not (scan_type == 'confocal' and
                    (laser_width != sensor_width or
                        laser_height != sensor_height)), \
            'If using scan_type=confocal, sensor_{width|height} must match laser_{width|height}'

        for y in range(laser_height):
            for x in range(laser_width):
                # start in (0, 1) space
                laser_lookat_x = (x + 0.5) / laser_width
                laser_lookat_y = (y + 0.5) / laser_height
                # take aperture into account
                laser_lookat_x = laser_aperture_start_x + \
                    laser_lookat_x * \
                    (laser_aperture_end_x - laser_aperture_start_x)
                laser_lookat_y = laser_aperture_start_y + \
                    laser_lookat_y * \
                    (laser_aperture_end_y - laser_aperture_start_y)
                # finally store in sensor space (0, sensor_width)
                laser_lookat_x *= sensor_width
                laser_lookat_y *= sensor_height
                laser_lookats.append((laser_lookat_x, laser_lookat_y))
    else:
        raise AssertionError(
            'Invalid scan_type, must be one of {single|exhaustive|confocal}')

    """ Create NLOSCaptureData and write all metadata """

    # TODO(diego): rotate sensor_grid_xyz and laser_grid_xyz based on relay wall rotation
    displacement = np.array([
        relay_wall['displacement_x'],
        relay_wall['displacement_y'],
        relay_wall['displacement_z']])
    sensor_grid_xyz = get_grid_xyz(
        sensor_width, sensor_height, relay_wall['scale_x'], relay_wall['scale_y'])
    sensor_grid_xyz += displacement
    if scan_type == 'single':
        px = relay_wall['scale_x'] * \
            ((laser_lookat_x / sensor_width) * 2 - 1)
        py = relay_wall['scale_y'] * \
            ((laser_lookat_y / sensor_height) * 2 - 1)
        laser_grid_xyz = np.array([[
            [px, py, 0],
        ]], dtype=np.float32)
    else:
        laser_grid_xyz = get_grid_xyz(
            laser_width, laser_height, relay_wall['scale_x'], relay_wall['scale_y'],
            ax0=laser_aperture_start_x, ax1=laser_aperture_end_x,
            ay0=laser_aperture_start_y, ay1=laser_aperture_end_y)
    laser_grid_xyz += displacement

    # TODO(diego): rotate [0, 0, 1] by rot_degrees_x (assmes RW is a plane)
    # or use a more generalist approach
    sensor_grid_normals = expand_xy_dims(
        np.array([0, 0, 1]), sensor_width, sensor_height)
    laser_grid_normals = expand_xy_dims(
        np.array([0, 0, 1]), laser_width, laser_height)

    capture_data = NLOSCaptureData()
    if scan_type == 'single' or scan_type == 'confocal':
        capture_data.H = np.zeros(
            (num_bins, sensor_width, sensor_height),
            dtype=np.float32)
        capture_data.H_format = HFormat.T_Sx_Sy
    elif scan_type == 'exhaustive':
        capture_data.H = np.zeros(
            (num_bins, laser_width, laser_height, sensor_width, sensor_height),
            dtype=np.float32)
        capture_data.H_format = HFormat.T_Lx_Ly_Sx_Sy
    else:
        raise AssertionError(
            'Invalid scan_type, must be one of {single|exhaustive|confocal}')
    capture_data.sensor_xyz = np.array([
        scene_config['sensor_x'],
        scene_config['sensor_y'],
        scene_config['sensor_z'],
    ], dtype=np.float32)
    capture_data.sensor_grid_xyz = sensor_grid_xyz
    capture_data.sensor_grid_normals = sensor_grid_normals
    capture_data.sensor_grid_format = GridFormat.X_Y_3
    capture_data.laser_xyz = np.array([
        scene_config['laser_x'],
        scene_config['laser_y'],
        scene_config['laser_z'],
    ], dtype=np.float32)
    capture_data.laser_grid_xyz = laser_grid_xyz
    capture_data.laser_grid_normals = laser_grid_normals
    capture_data.laser_grid_format = GridFormat.X_Y_3
    capture_data.delta_t = scene_config['bin_width_opl']
    capture_data.t_start = scene_config['start_opl']
    capture_data.t_accounts_first_and_last_bounces = \
        scene_config['account_first_and_last_bounces']
    capture_data.scene_info = {
        'tal_version': tal.__version__,
        'config': scene_config,
        'args': vars(args),
    }

    return scan_type, laser_lookats, capture_data


def __run_mitsuba(args, log_path, mitsuba_backend, scene_xml, hdr_path, defines, experiment_name, render_name, sensor_index,
                  check_done=lambda: False):
    if check_done():
        log(LogLevel.INFO, f'Skipping {render_name} for {experiment_name}')
        return

    class StdoutPipe:
        def __init__(self, pipe, logfile=None):
            self.pipe = pipe
            self.logfile = logfile

        def write(self, data):
            self.pipe.send(data)
            if logfile is not None:
                self.logfile.write(data)

        def close(self):
            self.pipe.send(None)
            self.pipe.close()
            if logfile is not None:
                self.logfile.close()

    pipe_r, pipe_w = multiprocessing.Pipe()
    logfile = None
    if args.do_logging and not args.dry_run:
        logfile = open(log_path, 'w')
    # NOTE: something here has a memory leak (probably Mitsuba-related)
    # We run Mitsuba in a separate process to ensure that the leaks do not add up
    # as they can fill your RAM in exhaustive scans
    run_mitsuba_f = partial(mitsuba_backend.run_mitsuba, scene_xml, hdr_path, defines,
                            render_name, args, StdoutPipe(pipe_w, logfile), sensor_index)
    if os.name == 'nt':
        # NOTE: Windows does not support multiprocessing
        run_mitsuba_f()
    else:
        process = multiprocessing.Process(target=run_mitsuba_f)
        try:
            process.start()
            while True:
                line = pipe_r.recv()
                if line:
                    if len(line.strip()) > 0:
                        log(LogLevel.INFO, line)
                else:
                    raise EOFError
        except EOFError:
            pipe_r.close()
            process.join()
        except KeyboardInterrupt:
            process.terminate()
            raise KeyboardInterrupt

    if args.do_logging and not args.dry_run:
        logfile.flush()
        logfile.close()


class RenderException(Exception):
    pass


def __merge_gt_results(args, mitsuba_backend, capture_data, gt_path):
    if not args.do_ground_truth_renders:
        return

    gt_image = mitsuba_backend.read_mitsuba_bitmap(gt_path)
    depth = gt_image[:, :, 0:3]
    normals = gt_image[:, :, 3:6]
    capture_data.scene_info['ground_truth'] = {
        'format': GroundTruthFormat.X_Y,
        'depth': depth,
        'normals': normals,
    }

    return capture_data


def __merge_nlos_results(args, mitsuba_backend, capture_data, partial_results_dir, experiment_name, scan_type, laser_lookats):

    if scan_type == 'single':
        hdr_path, _ = mitsuba_backend.partial_laser_path(
            partial_results_dir,
            experiment_name,
            *laser_lookats[0])
        capture_data.H = mitsuba_backend.read_transient_image(hdr_path)
    elif scan_type == 'exhaustive' or scan_type == 'confocal':
        laser_width = capture_data.H.shape[1]

        e_laser_lookats = enumerate(laser_lookats)
        if len(laser_lookats) > 1:
            e_laser_lookats = tqdm(
                e_laser_lookats, desc='Merging partial results...',
                file=TQDMLogRedirect(), ascii=True, total=len(laser_lookats))
        try:
            for i, (laser_lookat_x, laser_lookat_y) in e_laser_lookats:
                x = i % laser_width
                y = i // laser_width
                hdr_path, _ = mitsuba_backend.partial_laser_path(
                    partial_results_dir,
                    experiment_name,
                    laser_lookat_x, laser_lookat_y)
                if scan_type == 'confocal':
                    capture_data.H[:, x:x+1, y:y+1, ...] = \
                        mitsuba_backend.read_transient_image(hdr_path)
                elif scan_type == 'exhaustive':
                    capture_data.H[:, x, y, ...] = \
                        mitsuba_backend.read_transient_image(hdr_path)
                else:
                    raise AssertionError
        except Exception as exc:
            raise RenderException from exc
    else:
        raise AssertionError(
            'Invalid scan_type, must be one of {single|exhaustive|confocal}')

    return capture_data


def _main_render(config_path, args,
                 mitsuba_backend, scene_config,
                 root_dir, partial_results_dir, steady_dir, log_dir,
                 progress_file, num_retries=0):
    """ General initialization """

    steady_scene_xml, ground_truth_scene_xml, nlos_scene_xml = \
        __write_scene_xmls(args, mitsuba_backend, scene_config, root_dir)
    experiment_name = scene_config['name']

    """ Steady state + ground truth """

    if args.do_steady_renders:
        def render_steady(render_name, sensor_index):
            log(LogLevel.INFO,
                f'{render_name} for {experiment_name} steady render...')
            hdr_ext = mitsuba_backend.get_hdr_extension()
            hdr_path = os.path.join(partial_results_dir,
                                    f'{experiment_name}_{render_name}.{hdr_ext}')
            ldr_path = os.path.join(steady_dir,
                                    f'{experiment_name}_{render_name}.png')
            log_path = os.path.join(
                log_dir, f'{experiment_name}_{render_name}.log')

            __run_mitsuba(args, log_path, mitsuba_backend, steady_scene_xml, hdr_path, dict(),
                          experiment_name, render_name, sensor_index, check_done=lambda: os.path.exists(ldr_path))

            if not args.dry_run:
                mitsuba_backend.convert_hdr_to_ldr(hdr_path, ldr_path)

        render_steady('back_view', 0)
        render_steady('side_view', 1)

    if args.do_ground_truth_renders:
        gt_render_name = 'ground_truth'
        log(LogLevel.INFO, f'{gt_render_name} for {experiment_name}...')
        gt_ext = mitsuba_backend.get_hdr_extension()
        gt_path = os.path.join(partial_results_dir,
                               f'{experiment_name}_{gt_render_name}.{gt_ext}')
        log_path = os.path.join(
            log_dir, f'{experiment_name}_{gt_render_name}.log')

        __run_mitsuba(args, log_path, mitsuba_backend, ground_truth_scene_xml, gt_path, dict(),
                      experiment_name, gt_render_name, 0, check_done=lambda: os.path.exists(gt_path))

    """ NLOS renders """

    scan_type, laser_lookats, capture_data = \
        __write_metadata_and_get_laser_lookats(args, scene_config)

    pbar = tqdm(
        enumerate(laser_lookats), desc=f'Rendering {experiment_name} ({scan_type})...',
        file=TQDMLogRedirect(), ascii=True, total=len(laser_lookats))
    for i, (laser_lookat_x, laser_lookat_y) in pbar:
        try:
            hdr_path, is_dir = mitsuba_backend.partial_laser_path(
                partial_results_dir, experiment_name, laser_lookat_x, laser_lookat_y)
            if is_dir and not os.path.exists(hdr_path):
                os.mkdir(hdr_path)
        except OSError as exc:
            raise AssertionError(f'Invalid permissions: {exc}') from exc
        defines = {
            'laser_lookat_x': laser_lookat_x,
            'laser_lookat_y': laser_lookat_y,
        }
        log_path = os.path.join(
            log_dir,
            f'{experiment_name}_L[{laser_lookat_x}][{laser_lookat_y}].log')
        render_name = f'Laser {i + 1} of {len(laser_lookats)}'

        __run_mitsuba(args, log_path, mitsuba_backend, nlos_scene_xml, hdr_path, defines,
                      experiment_name, render_name, 0, check_done=lambda: os.path.exists(hdr_path))

        if scan_type == 'exhaustive' and i == 0:
            size_bytes = os.path.getsize(hdr_path)
            final_size_gb = size_bytes * len(laser_lookats) / 2**30
            pbar.set_description(
                f'Rendering {experiment_name} ({scan_type}, estimated size: {final_size_gb:.2f} GB)...')

    if args.dry_run:
        return

    """ Generate final HDF5 file"""

    log(LogLevel.INFO, 'Reading partial results and generating HDF5 file...')
    try:
        capture_data = __merge_gt_results(
            args, mitsuba_backend, capture_data, gt_path)
        capture_data = __merge_nlos_results(
            args, mitsuba_backend, capture_data, partial_results_dir, experiment_name, scan_type, laser_lookats)
    except RenderException:
        if num_retries >= 10:
            raise AssertionError(
                f'Failed to read partial results after {num_retries} retries')
        # TODO(diego): Mitsuba sometimes fails to write some images,
        # it seems like some sort of race condition
        # If there is a partial result missing, just re-launch for now
        mitsuba_backend.remove_transient_image(hdr_path)
        log(LogLevel.INFO,
            f'We missed some partial results (iteration {i} failed because: {exc}), re-launching...')
        return _main_render(config_path, args, num_retries=num_retries + 1)

    hdf5_path = os.path.join(root_dir, f'{experiment_name}.hdf5')
    tal.io.write_capture(hdf5_path, capture_data,
                         file_format=FileFormat.HDF5_TAL)
    log(LogLevel.INFO, f'Stored result in {hdf5_path}')

    if not args.keep_partial_results:
        log(LogLevel.INFO,
            f'Cleaning partial results in {partial_results_dir}...')
        shutil.rmtree(partial_results_dir, ignore_errors=True)
        log(LogLevel.INFO, f'All clean.')

    os.remove(progress_file)

    return hdf5_path


def render_nlos_scene(config_path, args):
    mitsuba_backend, scene_config, config_path = \
        _read_config_and_init_mitsuba_variant(config_path, args)

    root_dir, partial_results_dir, steady_dir, log_dir, progress_file = \
        _check_progress_and_create_folders(config_path, args)

    try:
        return _main_render(config_path, args,
                            mitsuba_backend, scene_config,
                            root_dir, partial_results_dir, steady_dir, log_dir,
                            progress_file)
    except KeyboardInterrupt:
        delete = None
        while delete is None:
            try:
                log(LogLevel.PROMPT,
                    f'Render cancelled. '
                    f'Delete the directory {root_dir}? (y/n): ', end='')
                answer = input()
                if answer.lower() == 'y':
                    delete = True
                elif answer.lower() == 'n':
                    delete = False
            except KeyboardInterrupt:
                pass
        if delete:
            shutil.rmtree(root_dir, ignore_errors=True)
            os.remove(progress_file)



from tal.log import log, LogLevel, TQDMLogRedirect

# pyright: reportMissingImports=false


def _get_setpath_location():
    from tal.config import ask_for_config, Config
    import os
    force_ask = False
    setpath_ok = False
    while not setpath_ok:
        setpath_location = os.path.join(
            ask_for_config(Config.MITSUBA2_TRANSIENT_NLOS_FOLDER,
                           force_ask=force_ask),
            'setpath.sh')
        if os.path.isfile(setpath_location):
            setpath_ok = True
        else:
            force_ask = True
            log(LogLevel.ERROR,
                f'setpath.sh cannot be found in {setpath_location}.', end='\n\n')
    return setpath_location


try:
    import mitsuba
except ModuleNotFoundError:
    import sys
    import subprocess
    setpath_location = _get_setpath_location()

    command = ['env', '-i', '/bin/bash',
               '-c', f'source {setpath_location} && printenv']
    p = subprocess.check_output(command).decode('utf-8')
    for line in p.split('\n')[:-1]:
        (key, _, value) = line.partition('=')
        if key == 'PYTHONPATH':
            for directory in value.split(':')[:-1]:
                sys.path.append(directory)


def get_name():
    return 'mitsuba2-transient-nlos'


def get_scene_version():
    return '2.2.1'


def get_long_version():
    return '2.2.1'


def get_default_variant():
    return 'scalar_rgb'


def set_variant(s):
    import mitsuba
    if not s.startswith('scalar_'):
        raise AssertionError(
            f'Variant {s} is not supported. It must start with "scalar_"')
    mitsuba.set_variant(s)


def get_hdr_extension():
    return 'exr'


def convert_hdr_to_ldr(hdr_path, ldr_path):
    from tal.util import write_img, tonemap_ldr
    image = read_mitsuba_bitmap(hdr_path)
    image = tonemap_ldr(image)
    write_img(ldr_path, image)


def partial_laser_path(partial_results_dir, experiment_name, lx, ly):
    import os
    return os.path.join(partial_results_dir, f'{experiment_name}_L[{lx}][{ly}]'.replace('.', '_')), True


def read_transient_image(path):
    return _read_mitsuba_streakbitmap(path)


def remove_transient_image(path):
    log(LogLevel.WARNING, f'Skipping the removal of {path}')
    return


def get_material_keys(s):
    import re
    return list(map(lambda e: e[1:],
                    re.compile(r'\$[a-zA-Z_]*').findall(s)))


def get_materials():
    from tal.util import fdent
    return {
        'white': fdent(f'''\
            <bsdf type="diffuse">
                <rgb name="reflectance" value="1.0, 1.0, 1.0"/>
            </bsdf>'''),
        'copper': fdent(f'''\
            <bsdf type="roughconductor">
                <string name="material" value="Cu"/>
                <string name="distribution" value="beckmann"/>
                <float name="alpha" value="$alpha"/>
            </bsdf>'''),
        'custom': '$text'
    }


def get_scene_xml(config, random_seed=0):
    import os
    import tal
    from tal.util import fdent

    def s(value):
        if isinstance(value, bool):
            return str(value).lower()
        else:
            return value

    def v(key):
        return s(config[key])

    # integrator
    integrator_steady = fdent(f'''\
        <integrator type="path"/>''')

    integrator_ground_truth = fdent(f'''\
        <integrator type="aov">
            <string name="aovs" value="pp:position, nn:geo_normal"/>
        </integrator>''')

    integrator_nlos = fdent(f'''\
        <integrator type="transientpath">
            <integer name="block_size" value="1"/>
            <integer name="max_depth" value="{v('integrator_max_depth')}"/>
            <integer name="filter_depth" value="{v('integrator_filter_depth')}"/>
            <boolean name="discard_direct_paths" value="{v('integrator_discard_direct_paths')}"/>
            <boolean name="nlos_laser_sampling" value="{v('integrator_nlos_laser_sampling')}"/>
            <boolean name="nlos_hidden_geometry_sampling" value="{v('integrator_nlos_hidden_geometry_sampling')}"/>
            <boolean name="nlos_hidden_geometry_sampling_do_rroulette" value="{v('integrator_nlos_hidden_geometry_sampling_do_rroulette')}"/>
            <boolean name="nlos_hidden_geometry_sampling_includes_relay_wall" value="{v('integrator_nlos_hidden_geometry_sampling_includes_relay_wall')}"/>
        </integrator>''')

    if 'polarized' in v('mitsuba_variant'):
        def add_stokes(s, itype):
            return fdent('''\
                <integrator type="{itype}">
                    {s}
                </integrator>''', itype=itype, s=s)
        integrator_steady = add_stokes(integrator_steady, 'stokes')
        integrator_nlos = add_stokes(integrator_nlos, 'transientstokes')

    # relay wall
    geometry_names = list(map(lambda g: g['name'], v('geometry')))
    assert len(geometry_names) == len(set(geometry_names)), \
        'One or more geometry names is duplicated'
    relay_wall_name = v('relay_wall')
    assert relay_wall_name in geometry_names, \
        f'Relay wall is set to {relay_wall_name}, but must be one of {geometry_names}'

    dummy_lights_and_geometry_steady = fdent(f'''\
        <!-- Colored spheres to mark the laser and sensor positions -->
        <shape type="sphere">
            <point name="center" x="{v('sensor_x')}" y="{v('sensor_y')}" z="{v('sensor_z')}"/>
            <float name="radius" value="0.05"/>
            <bsdf type="diffuse" id="red">
                <rgb name="reflectance" value="1.0, 0.0, 0.0"/>
            </bsdf>
        </shape>
        <shape type="sphere">
            <point name="center" x="{v('laser_x')}" y="{v('laser_y')}" z="{v('laser_z')}"/>
            <float name="radius" value="0.05"/>
            <bsdf type="diffuse" id="blue">
                <rgb name="reflectance" value="0.0, 0.0, 1.0"/>
            </bsdf>
        </shape>

        <!-- Illuminate all the scene -->
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="30" y="0" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="-30" y="0" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="0" y="30" z="30"/>
        </emitter>
        <emitter type="point">
            <rgb name="intensity" value="1.0, 1.0, 1.0"/>
            <point name="position" x="0" y="-30" z="30"/>
        </emitter>''')

    sensors_steady = fdent(f'''\
        <!-- Sensor 0: back view -->
        <sensor type="perspective">
            <string name="fov_axis" value="smaller"/>
            <float name="near_clip" value="0.01"/>
            <float name="far_clip" value="1000"/>
            <float name="fov" value="30"/>
            <transform name="to_world">
                <lookat origin="0, 0, 5"
                        target="0, 0, 0"
                            up="0, 1, 0"/>
            </transform>
            <sampler type="independent">
                <integer name="sample_count" value="512"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="512"/>
                <integer name="height" value="512"/>
                <rfilter name="rfilter" type="gaussian"/>
                <boolean name="high_quality_edges" value="false"/>
            </film>
        </sensor>

        <!-- Sensor 1: side view -->
        <sensor type="perspective">
            <string name="fov_axis" value="smaller"/>
            <float name="near_clip" value="0.01"/>
            <float name="far_clip" value="1000"/>
            <float name="fov" value="45"/>
            <transform name="to_world">
                <lookat origin="5, 0, 1.5"
                        target="0, 0, 1.5"
                            up="0, 1, 0"/>
            </transform>
            <sampler type="independent">
                <integer name="sample_count" value="512"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="512"/>
                <integer name="height" value="512"/>
                <rfilter name="rfilter" type="gaussian"/>
                <boolean name="high_quality_edges" value="false"/>
            </film>
        </sensor>''')

    confocal_capture = 'false'
    if v('scan_type') == 'confocal':
        confocal_capture = 'true'
    elif v('scan_type') != 'single' and v('scan_type') != 'exhaustive':
        raise AssertionError(
            'scan_type should be one of {single|confocal|exhaustive}')
    sensor_nlos = fdent(f'''\
        <sensor type="nloscapturemeter">
            <sampler type="independent">
                <integer name="sample_count" value="{v('sample_count')}"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>

            <emitter type="projector">
                <rgb name="irradiance" value="1.0, 1.0, 1.0"/>
                <float name="fov" value="{0.2 if v('integrator_nlos_laser_sampling') else 2}"/>
            </emitter>

            <boolean name="confocal" value="{confocal_capture}"/>
            <boolean name="account_first_and_last_bounces" value="{v('account_first_and_last_bounces')}"/>
            <point name="sensor_origin" x="{v('sensor_x')}" y="{v('sensor_y')}" z="{v('sensor_z')}"/>
            <point name="laser_origin" x="{v('laser_x')}" y="{v('laser_y')}" z="{v('laser_z')}"/>
            <point name="laser_lookat_pixel" x="$laser_lookat_x" y="$laser_lookat_y" z="0"/>
            <film type="streakhdrfilm" name="streakfilm">
                <integer name="width" value="{v('sensor_width')}"/>
                <integer name="height" value="{v('sensor_height')}"/>
                <string name="pixel_format" value="rgb"/>
                <string name="component_format" value="float32"/>

                <integer name="num_bins" value="{v('num_bins')}"/>
                <boolean name="auto_detect_bins" value="{v('auto_detect_bins')}"/>
                <float name="bin_width_opl" value="{v('bin_width_opl')}"/>
                <float name="start_opl" value="{v('start_opl')}"/>

                <rfilter name="rfilter" type="box"/>
                <!-- NOTE: tfilters are not implemented yet -->
                <!-- <rfilter name="tfilter" type="box"/>  -->
                <boolean name="high_quality_edges" value="false"/>
            </film>
        </sensor>''')

    materials = get_materials()
    shapes_steady = []
    shapes_ground_truth = []
    shapes_nlos = []
    for gdict in v('geometry'):
        def g(key):
            return s(gdict.get(key, None))

        name = g('name')
        is_relay_wall = name == relay_wall_name
        if is_relay_wall and g('mesh')['type'] != 'rectangle':
            log(LogLevel.WARNING, 'Relay wall does not work well with meshes that are '
                'not of type "rectangle" because of wrong UV mapping. '
                'Please make sure that you know what you are doing')
        shape_name = f'<!-- {name}{" (RELAY WALL)" if is_relay_wall else ""} -->'
        description = g('description')
        if description is not None and len(description.strip()) > 0:
            shape_name = fdent('''\
                {shape_name}
                <!--
                    {description}
                -->''', shape_name=shape_name, description=description.strip())

        material_id = g('material')['id']
        assert material_id in materials.keys(), \
            f'Geometry {name} has material {material_id}, but it must be one of {", ".join(materials.keys())}'
        shape_material = materials[material_id]
        material_keys = get_material_keys(shape_material)
        for key in material_keys:
            assert key in g('material'), \
                f'Material {material_id} for geometry {name} must have a value for {key}'
            shape_material = shape_material.replace(
                f'${key}', g('material')[key])

        shape_transform = fdent(f'''\
            <transform name="to_world">
                <scale x="{g('scale') or 1.0}" y="{g('scale') or 1.0}" z="{g('scale') or 1.0}"/>
                <rotate x="1" angle="{g('rot_degrees_x') or 0.0}"/>
                <rotate y="1" angle="{g('rot_degrees_y') or 0.0}"/>
                <rotate z="1" angle="{g('rot_degrees_z') or 0.0}"/>
                <translate x="{g('displacement_x') or 0.0}" y="{g('displacement_y') or 0.0}" z="{g('displacement_z') or 0.0}"/>
            </transform>''')

        shape_contents_steady = f'{shape_material}\n{shape_transform}'
        shape_contents_ground_truth = shape_contents_steady
        newline = '\n'
        shape_contents_nlos = f'{shape_material}\n{shape_transform}{f"{newline}{sensor_nlos}" if is_relay_wall else ""}'

        if is_relay_wall:
            # Transform on the ortographic camera depends on relay wall
            sensor_ground_truth_transform = fdent(f'''\
                <transform name="to_world">
                    <scale x="{g('scale_x') or 1.0}" y="{g('scale_y') or 1.0}"/>
                    <rotate x="1" angle="{g('rot_degrees_x') or 0.0}"/>
                    <rotate y="1" angle="{g('rot_degrees_y') or 0.0}"/>
                    <rotate z="1" angle="{g('rot_degrees_z') or 0.0}"/>
                    <translate x="{g('displacement_x') or 0.0}"
                               y="{g('displacement_y') or 0.0}"
                               z="{g('displacement_z') or 0.0}"/>
                </transform>''')

        if g('mesh')['type'] == 'obj':
            assert 'filename' in g('mesh'), \
                f'Missing mesh filename for geometry "{name}". ' \
                f'It is required because its mesh type is set as OBJ.'
            filename = g('mesh')['filename']
            assert os.path.isfile(filename), \
                f'{filename} does not exist for geometry "{name}"'

            def shapify(content, filename):
                return fdent('''\
                {shape_name}
                <shape type="obj">
                    <string name="filename" value="{filename}"/>

                    {content}
                </shape>''', shape_name=shape_name, filename=filename, content=content)

            shapes_steady.append(
                shapify(shape_contents_steady, filename))
            shapes_ground_truth.append(
                shapify(shape_contents_ground_truth, filename))
            shapes_nlos.append(
                shapify(shape_contents_nlos, filename))
        elif g('mesh')['type'] == 'rectangle' or g('mesh')['type'] == 'sphere':
            def shapify(content):
                return fdent('''\
                {shape_name}
                <shape type="{shape_type}">
                    {content}
                </shape>''', shape_name=shape_name, content=content,
                             shape_type=g('mesh')['type'])

            shapes_steady.append(
                shapify(shape_contents_steady, filename))
            shapes_ground_truth.append(
                shapify(shape_contents_ground_truth, filename))
            shapes_nlos.append(
                shapify(shape_contents_nlos, filename))
        else:
            raise AssertionError(
                f'Shape not yet supported: {g("mesh")["type"]}')

    shapes_steady = '\n\n'.join(shapes_steady)
    shapes_ground_truth = '\n\n'.join(shapes_ground_truth)
    shapes_nlos = '\n\n'.join(shapes_nlos)

    # Ground truth sensor declared here, after relay wall reading
    sensor_ground_truth = fdent('''\
        <!-- Ortographic camera (whose aperture corresponds to the relay wall) for depth and normals -->
        <sensor type="orthographic">
            {sensor_ground_truth_transform}
            <sampler type="independent">
                <integer name="sample_count" value="128"/>
                <integer name="seed" value="{random_seed}"/>
            </sampler>
            <film type="hdrfilm">
                <integer name="width" value="1028"/>
                <integer name="height" value="1028"/>
                <rfilter type="box">
                    <!-- <float name="radius" value="0.5"/> -->
                </rfilter>
            </film>
        </sensor>''',
                                sensor_ground_truth_transform=sensor_ground_truth_transform,
                                random_seed=random_seed)

    file_steady = fdent('''\
        <!-- Auto-generated using TAL v{tal_version} -->
        <scene version="{mitsuba_version}">
            {integrator_steady}

            {dummy_lights_and_geometry_steady}

            {shapes_steady}

            {sensors_steady}
        </scene>''',
                        tal_version=tal.__version__,
                        mitsuba_version=get_scene_version(),
                        integrator_steady=integrator_steady,
                        dummy_lights_and_geometry_steady=dummy_lights_and_geometry_steady,
                        shapes_steady=shapes_steady,
                        sensors_steady=sensors_steady)

    file_ground_truth = fdent('''\
        <!-- Auto-generated using TAL v{tal_version} -->
        <scene version="{mitsuba_version}">
            {integrator_ground_truth}

            {shapes_ground_truth}

            {sensor_ground_truth}
        </scene>''',
                              tal_version=tal.__version__,
                              mitsuba_version=get_scene_version(),
                              integrator_ground_truth=integrator_ground_truth,
                              shapes_ground_truth=shapes_ground_truth,
                              sensor_ground_truth=sensor_ground_truth)

    file_nlos = fdent('''\
        <!-- Auto-generated using TAL v{tal_version} -->
        <scene version="{mitsuba_version}">
            {integrator_nlos}

            {shapes_nlos}
        </scene>''',
                      tal_version=tal.__version__,
                      mitsuba_version=get_scene_version(),
                      integrator_nlos=integrator_nlos,
                      shapes_nlos=shapes_nlos)

    return file_steady, file_ground_truth, file_nlos


def run_mitsuba(scene_xml_path, hdr_path, defines,
                experiment_name, logfile, args, sensor_index=0, queue=None):
    import re
    import time
    import subprocess
    import os
    from tqdm import tqdm
    # execute mitsuba command (sourcing setpath.sh before)
    num_threads = args.threads
    command = ['mitsuba',
               '-o', hdr_path,
               '-s', str(sensor_index),
               '-t', str(num_threads)]
    for key, value in defines.items():
        command += ['-D', f'{key}={value}']
    command += [scene_xml_path]

    nice = args.nice
    command = ['nice', '-n', str(nice), " ".join(command)]

    setpath_location = _get_setpath_location()

    if args.dry_run:
        # add extra commas to make it easier to copy-paste
        command = ['/bin/bash', '-c',
                   f'"source \\"{setpath_location}\\" && {" ".join(command)}"']
        log(LogLevel.PROMPT, ' '.join(command))
        return
    else:
        command = ['/bin/bash', '-c',
                   f'source "{setpath_location}" && {" ".join(command)}']

    # need to pass the command through stdbuf to be able to read the progress bar
    command = ['stdbuf', '-o0'] + command
    mitsuba_process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    # read the progress bar and pass the info to the user through a tqdm bar
    # this is totally not overengineering-trust me-this saves so much time
    progress_re = re.compile(
        r'Rendering \[(=* *)\] \([\d\.]+\w+, ETA: ([\d\.]+\w+)\)')
    read_opl = defines.get('auto_detect_bins', False)
    if read_opl:
        opl_output = ''
        opl_re = re.compile(
            r'limits: \[(\d+\.\d+), \d+\.\d+\] with bin width (\d+\.\d+)')
    with tqdm(desc=experiment_name, total=100, ascii=True, leave=False,
              file=TQDMLogRedirect(),
              bar_format='{desc} |{bar}| [{n:.2f}%{postfix}] ') as pbar:
        output = None
        while output is None or len(output) > 0:
            output = mitsuba_process.stdout.read(160)
            try:
                output = output.decode('utf-8')
            except UnicodeDecodeError:
                continue
            if logfile is not None:
                logfile.write(output)
                logfile.flush()
            if read_opl:
                opl_output += output
                matches = opl_re.findall(opl_output)
                if len(matches) > 0:
                    start_opl, bin_width_opl = matches[-1]
                    start_opl = float(start_opl)
                    bin_width_opl = float(bin_width_opl)
                    log(LogLevel.INFO, 'Auto-detected histogram: '
                        f'start_opl={start_opl:.4f}, bin_width_opl={bin_width_opl:.6f}')
                    defines.update(start_opl=start_opl)
                    defines.update(bin_width_opl=bin_width_opl)
                    read_opl = False
                    del opl_output
            matches = progress_re.findall(output)
            if len(matches) > 0:
                progress, eta = matches[-1]
                completed = progress.count('=')
                not_completed = progress.count(' ')
                progress = 100 * completed / (completed + not_completed)
                pbar.update(progress - pbar.n)
                pbar.set_postfix_str(f'ETA: {eta}')
                if not_completed == 0:
                    break
                time.sleep(1)

    # wait for the process to end
    mitsuba_process.communicate()
    assert mitsuba_process.returncode == 0, \
        f'Mitsuba returned with error code {mitsuba_process.returncode}'
    if os.path.isdir(hdr_path):
        # assume transient render
        n_exr_frames = len(os.listdir(hdr_path))
        assert n_exr_frames > 0, 'No frames were rendered?'
    else:
        # assume steady state render
        assert os.path.isfile(hdr_path), 'No frames were rendered?'


def read_mitsuba_bitmap(path: str):
    from mitsuba.core import Bitmap
    import numpy as np
    return np.array(Bitmap(path), copy=False)


def _read_mitsuba_streakbitmap(path: str, exr_format='RGB'):
    """
    Reads all the images x-t that compose the streak image.

    :param dir: path where the images x-t are stored
    :return: a streak image of shape [time, width, height]
    """
    import re
    import glob
    import os
    import numpy as np
    from tqdm import tqdm

    # NOTE(diego): for now this assumes that the EXR that it reads
    # are in RGB format, and returns an image with 3 channels,
    # in the case of polarized light it can return something else
    if exr_format != 'RGB':
        raise NotImplementedError(
            'Formats different from RGB are not implemented')

    xtframes = glob.glob(os.path.join(
        glob.escape(path), f'frame_*.exr'))
    xtframes = sorted(xtframes,
                      key=lambda x: int(re.compile(r'\d+').findall(x)[-1]))
    number_of_xtframes = len(xtframes)
    first_img = read_mitsuba_bitmap(xtframes[0])
    streak_img = np.empty(
        (number_of_xtframes, *first_img.shape), dtype=first_img.dtype)
    with tqdm(desc=f'Reading {path}', total=number_of_xtframes, file=TQDMLogRedirect(), ascii=True) as pbar:
        for i_xtframe in range(number_of_xtframes):
            other = read_mitsuba_bitmap(xtframes[i_xtframe])
            streak_img[i_xtframe] = np.nan_to_num(other, nan=0.)
            pbar.update(1)

    # for now streak_img has dimensions (y, x, time, channels)
    assert streak_img.shape[-1] == 3, \
        f'Careful, streak_img has shape {streak_img.shape} (i.e. its probably not RGB as we assume, last dimension should be 3)'
    # and we want it as (time, x, y)
    return np.sum(np.transpose(streak_img), axis=0)



from enum import Enum
import os
import h5py
import yaml
import numpy as np
import inspect
from typing import Union, get_type_hints
from nptyping import NDArray, Shape
from tal.io.format import convert_dict, detect_dict_format
from tal.enums import FileFormat, HFormat, GridFormat, VolumeFormat
from tal.log import log, LogLevel


class LazyDataset:
    def __init__(self, dataset):
        self.read = False
        self.dataset = dataset

    def __get__(self, instance, owner):
        if not self.read:
            self.dataset = self.dataset[()]
            self.read = True
        return self.dataset

    def __getattr__(self, name):
        return getattr(self.dataset, name)


def read_hdf5(filename: str) -> dict:
    raw_data = h5py.File(filename, 'r')

    def parse(key, value):
        if isinstance(value, h5py.Empty):
            value = None
        elif isinstance(value, h5py.Group):
            value = {k: parse(k, v) for k, v in value.items()}
        else:
            if isinstance(value, h5py.Dataset):
                value = value[()]
                # FIXME(diego): unused, need to figure out how
                # to call do stuff like value.size or
                # isinstance(value, bytes) without reading the
                # whole dataset
                # value = LazyDataset(value)
            if isinstance(value, np.ndarray) and value.size == 1:
                value = np.squeeze(value)[()]

        # parse enums and strings
        if key in get_type_hints(NLOSCaptureData):
            key_class = get_type_hints(NLOSCaptureData)[key]
            if inspect.isclass(key_class) and issubclass(key_class, Enum):
                if isinstance(value, h5py.Empty):
                    value = key_class(0)
                else:
                    value = key_class(value)
        elif isinstance(value, bytes):
            # FIXME(diego): not working?
            value = yaml.safe_load(value)
        return value

    raw_data = dict((key, parse(key, raw_data.get(key)))
                    for key in raw_data.keys())
    return raw_data


def write_hdf5(filename: str, capture_data: dict):
    file = h5py.File(filename, 'w')
    for key, value in capture_data.items():
        if value is None:
            file[key] = h5py.Empty(float)
        elif isinstance(value, dict):
            file[key] = yaml.dump(value, Dumper=yaml.CDumper)
        elif isinstance(value, Enum):
            dt = h5py.enum_dtype(dict((item.name, item.value)
                                 for item in value.__class__), basetype='i')
            ds = file.create_dataset(key, (1,), dtype=dt)
            ds[0] = value.value
        else:
            file[key] = value
    file.close()


class NLOSCaptureData:
    """
    Contains data for a NLOS capture scene. Typically, a laser illuminates a planar wall with a delta light pulse.
    An ultrafast camera captures the reflected light at that wall.

    The scene is described by the following parameters:

    H
        Multi-dimensional (>=2 dimensions) impulse response function.
        Contains time-resolved illumination for different laser and sensor positions.
        See H_format for the number of dimensions and what each one contains.

    H_format
        Enum describing what each dimension contains.
        e.g. HFormat.T_Lx_Ly_Sx_Sy means that there is temporal (T) data for a 2D grid
             of laser positions (Lx, Ly) and a 2D grid of sensor positions (Sx, Sy)
        e.g. HFormat.T_Sx_Sy means that there is temporal (T) data for a 2D grid of sensor positions.
             In this case, see is_confocal() to differentiate between the case of only one laser point
             and the case of a 2D grid of laser positions co-located with the sensor positions.
             You can also check sensor_grid_xyz and laser_grid_xyz, but is_confocal() is easier.

    sensor_xyz, laser_xyz
        3D coordinates of the sensor and laser positions.
        Note that it is the origin of the laser itself, and _not_ the point where the laser is focused at.

    sensor_grid_xyz, laser_grid_xyz
        Positions of laser and sensor points at the wall.
        See {sensor|laser}_grid_format for the number of dimensions.

    sensor_grid_normals, laser_grid_normals
        Normal vectors of the laser and sensor points at the wall.
        See {sensor|laser}_grid_format for the number of dimensions.

    sensor_grid_format, laser_grid_format
        Enum describing what each dimension of {sensor|laser}_grid_{xyz|normals} contains.
        See tal.enums.GridFormat for details.

    delta_t
        Temporal resolution of H i.e. distance between two time bins.
        Measured in meters (see https://en.wikipedia.org/wiki/Optical_path_length).

    t_start
        Initial instant of H, used to define a different time origin.
        Measured in meters (see https://en.wikipedia.org/wiki/Optical_path_length).

    t_accounts_first_and_last_bounces
        Whether the impulse response accounts for the time-of-flight between
        the laser origin and the wall (first bounce),
        and the wall and the sensor origin (last bounce).

    scene_info
        YAML-encoded string. Contains additional information about the scene. Implemented keys:
        - 'original_format': str (e.g. 'HDF5_ZNLOS' if imported from ZNLOS dataset instead of tal render)
        - 'config': dict (original scene_config.yaml when generated using tal render)
        - 'args': dict (original args passed when generated using tal render)
        - 'volume': dict (if available e.g. for 'HDF5_ZNLOS' datasets, *not for tal render*)
            - 'center': Array3 (center of volume)
            - 'rotation': Array3 (rotation of volume formatted as per Z-NLOS - probably unused for now)
            - 'size': Array3 (size of volume - length of each XYZ side)
            - 'xyz': MatrixN3 (if ever volume_format is VolumeFormat.N_3, points will be stored here)
        - 'ground_truth': dict (if available, e.g. generated by tal render).
            Remember that TAL uses a XYZ coordinate system where the relay wall is a plane at Z = 0,
            and Z > 0 is the depth axis away from the relay wall.
            All ground truth values use (X, Y) coordinates for the relay wall plane
            - 'format': GroundTruthFormat (for now it is GroundTruthFormat.X_Y always)
            - 'depth': GroundTruthDepthType (TensorXY)
                Example: depth[i, j] stores the distance to the hidden geometry,
                    from the relay wall to the hit point.
                It has size 1024 x 1024. depth[0, 0] corresponds to the top-left
                    of the relay wall, depth[-1, -1] corresponds to the bottom-right of the relay wall.
                If it does not hit the hidden geometry, depth[i, j] = -1
                If it hits, depth[i, j] stores the distance from the relay wall to the hit point.
            - 'normals': GroundTruthNormalsType (TensorXY3)
                Similar to depth, but stores the normal of the hidden geometry at the hit point.
    """

    #
    # Type aliases
    #
    Float = np.float32
    TensorTSxSy = NDArray[Shape['T, Sx, Sy'], Float]
    TensorTLxLySxSy = NDArray[Shape['T, Lx, Ly, Sx, Sy'], Float]
    HType = Union[TensorTSxSy, TensorTLxLySxSy]
    MatrixN3 = NDArray[Shape['*, 3'], Float]
    TensorXY3 = NDArray[Shape['X, Y, 3'], Float]
    LaserGridType = Union[MatrixN3, TensorXY3]
    SensorGridType = Union[MatrixN3, TensorXY3]
    TensorXYZ3 = NDArray[Shape['X, Y, Z, 3'], Float]
    VolumeXYZType = Union[MatrixN3, TensorXY3, TensorXYZ3]
    Array3 = NDArray[Shape['3'], Float]

    # reconstruction types
    TensorN = NDArray[Shape['N'], Float]
    TensorXY = NDArray[Shape['X, Y'], Float]
    TensorXYZ = NDArray[Shape['X, Y, Z'], Float]
    SingleReconstructionType = Union[TensorN, TensorXY, TensorXYZ]
    TensorNN = NDArray[Shape['N, N'], Float]
    TensorXYXY = NDArray[Shape['X, Y, X, Y'], Float]
    TensorXYZXYZ = NDArray[Shape['X, Y, Z, X, Y, Z'], Float]
    ExhaustiveReconstructionType = Union[TensorNN, TensorXYXY, TensorXYZXYZ]
    GroundTruthDepthType = TensorXY
    GroundTruthNormalsType = TensorXY3

    #
    # Actual capture data (ignore _start and _end)
    #
    _start: None = None  # used in as_dict()
    H: HType = None
    H_format: HFormat = None
    sensor_xyz: Array3 = None
    sensor_grid_xyz: SensorGridType = None
    sensor_grid_normals: SensorGridType = None
    sensor_grid_format: GridFormat = None
    laser_xyz: Array3 = None
    laser_grid_xyz: LaserGridType = None
    laser_grid_normals: LaserGridType = None
    laser_grid_format: GridFormat = None
    # volume_xyz : VolumeXYZType = None  # deprecated
    volume_format: VolumeFormat = None  # deprecated
    delta_t: Float = None
    t_start: Float = None
    t_accounts_first_and_last_bounces: bool = None
    scene_info: dict = None  # additional information
    _end: None = None  # used in as_dict()

    def __get_dict_keys(self):
        variables = list(self.__class__.__dict__.keys())
        variables = variables[variables.index(
            '_start') + 1: variables.index('_end')]
        return variables

    def __init__(self, filename: str = None, file_format: FileFormat = FileFormat.AUTODETECT):
        if filename is None:
            return

        # Read raw data and check its current format
        final_path = filename
        while os.path.islink(final_path):
            aux_filename = final_path
            assert os.path.exists(aux_filename), \
                f'File {aux_filename} is a broken link'
            # Access the original data in case it is a link
            log(LogLevel.DEBUG, f'{aux_filename} is a link. Following link...')
            aux_filename = os.readlink(aux_filename)
            final_path = os.path.join(
                os.path.dirname(final_path), aux_filename)

        assert os.path.isfile(final_path), f'Does not exist: {final_path}'
        raw_data = read_hdf5(final_path)
        if file_format == FileFormat.AUTODETECT:
            file_format = detect_dict_format(raw_data)

        # Convert data to HDF5_TAL format
        if file_format != FileFormat.HDF5_TAL:
            raw_data = convert_dict(
                raw_data, format_to=FileFormat.HDF5_TAL)

        own_dict_keys = self.__get_dict_keys()
        for key, value in raw_data.items():
            if key not in own_dict_keys:
                raise AssertionError(f'raw_data contains unknown key: {key}')
            if key == 'scene_info' and not isinstance(value, h5py.Empty):
                value = yaml.load(value, Loader=yaml.CLoader)

            setattr(self, key, value)

    def is_laser_paired_to_sensor(self):
        """
        Returns True if each sensor point only has one laser point associated.
        This is the case for confocal capture or custom configurations.
        In single or exhaustive configuration, this does not happen as each
        laser/sensor point can have multiple sensor/laser points associated.
        Some reconstructions algorithms require this information to work properly.
        """
        if self.H_format in [HFormat.T_Lx_Ly_Sx_Sy, HFormat.T_Li_Si]:
            return False
        elif self.H_format in [HFormat.T_Sx_Sy, HFormat.T_Si]:
            return (self.sensor_grid_format == self.laser_grid_format
                    and
                    self.sensor_grid_xyz.shape[:-1] ==
                    self.laser_grid_xyz.shape[:-1])
        else:
            raise AssertionError('Invalid H_format')

    def is_confocal(self):
        """ Returns True if H contains confocal data (i.e. Sx and Sy represent both laser and sensor coordinates) """
        if self.H_format in [HFormat.T_Lx_Ly_Sx_Sy, HFormat.T_Li_Si]:
            return False
        elif self.H_format in [HFormat.T_Sx_Sy, HFormat.T_Si]:
            return np.allclose(self.sensor_grid_xyz, self.laser_grid_xyz)
        else:
            raise AssertionError('Invalid H_format')

    def as_dict(self):
        """ Returns a dict containing all the data in this object """
        dict_keys = self.__get_dict_keys()
        return dict((key, getattr(self, key)) for key in dict_keys)

    def downscale(self, downscale: int):
        """
        Updates the data in this object to reduce the number of laser and sensor positions by the given factor

        e.g. Consider data with H_format = (T, Sx, Sy) and H.shape = (4096, 256, 256).
             Calling with downscale = 2 will reduce H.shape to (4096, 128, 128).
        """
        assert downscale > 1, 'downscale must be > 1'
        assert self.H_format in [HFormat.T_Sx_Sy], \
            'Only implemented for HFormat.T_Sx_Sy'
        nt, nsx, nsy = self.H.shape
        self.H = self.H.reshape(
            (nt, nsx // downscale, downscale, nsy // downscale, downscale)).sum(axis=(2, 4))
        self.sensor_grid_xyz = self.sensor_grid_xyz.reshape(
            (nsx // downscale, downscale, nsy // downscale, downscale, 3)).mean(axis=(1, 3))
        log(LogLevel.INFO,
            f'Downscaled from {nsx}x{nsy} to {nsx // downscale}x{nsy // downscale}')

    def get_single_subdata_from_laser_point(self, *args):
        """
        If your data has more than one illumination point, you can use this function
        to obtain the data for a single illumination point.

        The args specify the indices of the laser point in the same format as
        the impulse response H or the laser positions laser_grid_xyz.

        e.g. If H_format = (T, Lx, Ly, Sx, Sy), then args = (5, 10) will return
             the impulse response at (T, 5, 10, Sx, Sy).
        """
        n_indices = len(args)

        if n_indices == 1:
            assert self.H_format == HFormat.T_Li_Si and self.laser_grid_format == GridFormat.N_3, \
                'Number of indices must match H_format and laser_grid_format'
            il = args[0]
            assert 0 <= il and \
                il < self.laser_grid_xyz.shape[0] and \
                il < self.laser_grid_normals.shape[0] and \
                il < self.H.shape[1], \
                'Index out of bounds'
            slices = np.index_exp[il:il+1]
        elif n_indices == 2:
            assert self.H_format == HFormat.T_Lx_Ly_Sx_Sy and self.laser_grid_format == GridFormat.X_Y_3, \
                'Number of indices must match H_format and laser_grid_format'
            ilx, ily = args
            assert 0 <= ilx and 0 <= ily and \
                ilx < self.laser_grid_xyz.shape[0] and \
                ily < self.laser_grid_xyz.shape[1] and \
                ilx < self.laser_grid_normals.shape[0] and \
                ily < self.laser_grid_normals.shape[1] and \
                ilx < self.H.shape[1] and \
                ily < self.H.shape[2], \
                'Index out of bounds'
            slices = np.index_exp[ilx:ilx+1, ily:ily+1]
        else:
            raise AssertionError('Invalid number of indices')

        from copy import deepcopy
        data_out = deepcopy(self)
        # cannot use * operator directly inside slice [...]
        # as it is not supported until python 3.11
        # for now it's rewritten by wrapping everything inside a tuple
        data_out.H = data_out.H[(slice(None), *slices, ...)]
        data_out.laser_grid_xyz = \
            data_out.laser_grid_xyz[(*slices, ...)]
        data_out.laser_grid_normals = \
            data_out.laser_grid_normals[(*slices, ...)]

        return data_out
